/*
 * The contents of this file are subject to the Mozilla Public License
 * Version 1.1 (the "License");  you may not use this file except in compliance
 * with the License.  You may obtain a copy of the License at
 * http://www.mozilla.org/MPL/
 *
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License for
 * the specific language governing rights and limitations under the License.
 *
 * The Original Code is Protege-2000.
 *
 * The Initial Developer of the Original Code is Stanford University. Portions
 * created by Stanford University are Copyright (C) 2001.  All Rights Reserved.
 *
 * Protege-2000 was developed by Stanford Medical Informatics
 * (http://www.smi.stanford.edu) at the Stanford University School of Medicine
 * with support from the National Library of Medicine, the National Science
 * Foundation, and the Defense Advanced Research Projects Agency.  Current
 * information about Protege can be obtained at http://protege.stanford.edu
 *
 * This code developed by The Department of Biomedical and Health Informatics at 
 * at the University of Washington (http://www.bhi.washington.edu)
 *
 */
 
package edu.washington.bhi.protegex.storage.xml;

//import util.Arguments;

import java.io.*;
import java.util.*;

import org.w3c.dom.*;

import edu.stanford.smi.protege.model.*;
import edu.stanford.smi.protege.resource.*;

/* Reads XML into Protege
 */
 
public class ProtegeXMLImporter extends XMLImporter{
	
	private static int MIN_TRACE_COUNT = 5000;
    private static int TRACE_COUNT_DELTA = 1000;
    private static int MIN_CLASS_TRACE_COUNT = MIN_TRACE_COUNT;
    private static int MIN_INSTANCE_TRACE_COUNT = MIN_TRACE_COUNT;
    private static int CLASS_TRACE_COUNT_DELTA = TRACE_COUNT_DELTA;
    private static int INSTANCE_TRACE_COUNT_DELTA = TRACE_COUNT_DELTA;
    private Collection itsUndefinedInstances = new HashSet();
    private Collection itsUndefinedClses = new HashSet();
    private boolean localIsIncluded;
    private String state;
	
	protected KnowledgeBase mainkb;
    
    private Set _existingStdClasses = new HashSet();

    {
       _existingStdClasses.add(Model.Cls.STANDARD_CLASS);
       _existingStdClasses.add(Model.Cls.THING);
    }
    
    private Set _existingClasses = new HashSet();
	
	
	public ProtegeXMLImporter(KnowledgeBase kb, boolean isIncluded) {
		mainkb = kb;
		localIsIncluded = isIncluded;
	}
	
	
	public void parseDocument(Node mainnode) {
		
		//get the ontology name
		Node child = mainnode.getFirstChild();
		String name = child.getNodeName();
		
		for(child = child.getFirstChild(); child != null; child = child.getNextSibling()) {
	
	        int type = child.getNodeType();
	        switch(type) {
	            case Node.DOCUMENT_NODE: {  
	              break;                  
	            }//end of document
	            case Node.ELEMENT_NODE: {
					name = child.getNodeName().trim();

					
					if (name.equals("classes")) {
						parseClasses(child);
					}
					
					else if (name.equals("slots")) {
						parseSlots(child);
					}
					
					else if (name.equals("instances")) {
						parseInstances(child);
					}
		   		}
		   		case Node.TEXT_NODE: {
		   		}	   		
			}
		}
	}	
	
	protected void parseClasses(Node node) {
		Node child;
		Hashtable class_node_hash = new Hashtable();
		Hashtable class_sclass_hash = new Hashtable();
				
		for(child = node.getFirstChild(); child != null; child = child.getNextSibling()) {
	        int type = child.getNodeType();
	        switch(type) {
	            case Node.DOCUMENT_NODE: {  
	              break;                  
	            }//end of document
	            case Node.ELEMENT_NODE: {
					String name = child.getNodeName().trim();
					
					if (name.equals("class")) {

						//take a first parse through. if superclass exists, then ok, else stick in 
						//hash for future passes
						 
						//get superclasses for this class 
						//get metacls for this class
						
						Collection superMetaClsList = getSuperMetaCls(child);
						
						
						//if superclass and metacls exist then go ahead
						if (existsClasses(superMetaClsList)) {
							parseClass(child);
						}
						else {
							
							// get the name of the child 
							String clsName = getName(child);
							
							//else put it into the hash and move onto the next one
							class_node_hash.put(clsName, child);
							class_sclass_hash.put(clsName, superMetaClsList);
						}
					}
		   		}	   		
			}
		}
		
		
		// Now loop through the hashes and until done and add classes
		// print to screen list of classes that didn't make it
		
		//mark false if actually added a class so can look for more subclasses
		boolean done = false;
		
		while (!done) {
			
			done = true;
				
			for ( Enumeration e = class_node_hash.keys() ; e.hasMoreElements();) {
				// retrieve the class name
				String thisCls = (String) e.nextElement();
	
				// check its superclasses
				Collection sclasses = (Collection)class_sclass_hash.get(thisCls);
				
				if (existsClasses(sclasses)) {
					//add it
					parseClass((Node)class_node_hash.get(thisCls));
	
					//delete from hash
					class_node_hash.remove(thisCls);
					class_sclass_hash.remove(thisCls);
					
					//say found one
					done = false;
				}
			}
		}	
		
		printExtraClasses(class_node_hash);
	}
	
	
	private boolean existsClasses(Collection classes) {
		boolean classesExist = true;
		
		Iterator i = classes.iterator();
		while(i.hasNext()){
			String sClass = (String)i.next();			
			if (!doesClassExist(sClass)) {
				classesExist = false;
				return classesExist;
			}
		}
		return classesExist;
	}
	
	
	//gets the name of a class
 	private String getName(Node node) {
		String name = "";
		String nodename;		
		Node child, textnode;
		
		child = node.getFirstChild();

		for(child = child.getFirstChild(); child != null; child = child.getNextSibling()) {
	  		nodename = getSlotName(child);
	  		if (nodename.equals(Model.Slot.NAME)) {
	  			textnode = child.getFirstChild();
	  			name = getTextnodevalue(textnode);
	  		}
	  	}
	  	
	  	return name;
	}
	
	
		private Collection getSuperMetaCls(Node node) {
		Node child, child2;
//		Collection superClasses = new HashSet();
//		Collection directType = new HashSet();
		Collection superMetaCls = new HashSet();
		String cls;
		
		//should be getting to "own-slots"
		child = node.getFirstChild();
		
		//should be getting to "own-slot"
//		child = child.getFirstChild();
		
	  	for(child = child.getFirstChild(); child != null; child = child.getNextSibling()) {
	  		String nodeName = getSlotName(child);
	  		
	  		if (nodeName.equals(Model.Slot.DIRECT_SUPERCLASSES)) {
	  			
//	  			superClasses = parseOwnSlot(child);
	  			
	  			for(child2 = child.getFirstChild(); child2 != null; child2 = child2.getNextSibling()) {	  			
					//get values for superclasses
					cls = parseClsString(getTextnodevalue(child2.getFirstChild()));
					if (!cls.equals("") && !_existingStdClasses.contains(cls)) {
			    	    superMetaCls.add(cls);
			    	}
				}
		  	}

		  	else if (nodeName.equals(Model.Slot.DIRECT_TYPES)) {
		  		
//		  		directType = parseOwnSlot(child);
		  		
		  		child2 = child.getFirstChild();
		  		
	  			cls = parseClsString(getTextnodevalue(child2.getFirstChild()));
	  			if (!cls.equals("") && !_existingStdClasses.contains(cls)) {
    		        superMetaCls.add(cls);
    		    }
				
		  	}
	  	}
	  	
	  	
//	  	Iterator i = directType.iterator();
//		while(i.hasNext()){
//			superClasses.add(i.next());
//		}
	  	
		return superMetaCls;		
	}
	
	private String getSlotName(Node node) {
		String slotName = getAttribute(node, "slot-name");	        
        return slotName;	
	}
	
	
	protected void parseSlots(Node node) {
		Node child;
				
		for(child = node.getFirstChild(); child != null; child = child.getNextSibling()) {
	        int type = child.getNodeType();
	        switch(type) {
	            case Node.DOCUMENT_NODE: {  
	              break;                  
	            }//end of document
	            case Node.ELEMENT_NODE: {
					String name = child.getNodeName().trim();
					
					if (name.equals("slot")) {
						parseSlot(child);
					}
		   		}	   		
			}
		}
	}
	
	
	protected void parseInstances(Node node) {
		Node child;
				
		for(child = node.getFirstChild(); child != null; child = child.getNextSibling()) {
	        int type = child.getNodeType();
	        switch(type) {
	            case Node.DOCUMENT_NODE: {  
	              break;                  
	            }//end of document
	            case Node.ELEMENT_NODE: {
					String name = child.getNodeName().trim();
					
					if (name.equals("instance")) {
						parseInstance(child);
					}
		   		}	   		
			}
		}
	}
	
	private Object getSingleValue(Node node) {
		Node child, child2;
		Object obj = new Object();
		
		child = node.getFirstChild();
		
		if (child.getNodeName().equals("entry")) {
				ValueType type = ValueType.valueOf(getAttribute(child, "type"));
				child2 = child.getFirstChild();
				
				if (child2.getNodeName().equals("value")) {
					String value = getTextnodevalue(child2);
					obj = convertValue(type, value);
				}
		}
		
		return obj;	
	}
	
	//returns Collection with correctly typed objects
	protected Collection parseOwnSlot(Node node) {
		Collection slotColl = new HashSet();
		Node child, child2;
		String value;

	 	for(child = node.getFirstChild(); child!= null; child = child.getNextSibling()) {					
			if (child.getNodeName().equals("entry")) {
				ValueType type = ValueType.valueOf(getAttribute(child, "type"));
				for(child2 = child.getFirstChild(); child2!= null; child2 = child2.getNextSibling()) {
					if (child2.getNodeName().equals("value")) {
						value = getTextnodevalue(child2);
						Object obj = convertValue(type, value);
						slotColl.add(obj);
					}
				}			

			}
		}
		
			return slotColl;
		
	}
	

	//returns Collection with correctly typed objects
	protected ArrayList parseValueType(Node node) {
		ArrayList slotArray = new ArrayList();
		Node child, child2;
		String value;

	 	for(child = node.getFirstChild(); child!= null; child = child.getNextSibling()) {					
			if (child.getNodeName().equals("entry")) {
				ValueType type = ValueType.valueOf(getAttribute(child, "type"));
				for(child2 = child.getFirstChild(); child2!= null; child2 = child2.getNextSibling()) {
					if (child2.getNodeName().equals("value")) {
						value = getTextnodevalue(child2);
						Object obj = convertValue(type, value);
						slotArray.add(obj);
					}
				}			
			}
		}

		return slotArray;
		
	}
	
	protected void parseClass(Node node) {
		//the new class
		Cls cls;
		Node child, child2, child3, child4, textnode;
		String clsName, value;	
		
		child = node.getFirstChild();
		child2 = child.getFirstChild();
		
		if (!getSlotName(child2).equals(Model.Slot.NAME)) {
			//oops
			printErrorMessage("First node is not " + Model.Slot.NAME);
			return;
		}
		else {
			clsName = (String) getSingleValue(child2);				
			cls = getClass(clsName);
			_existingClasses.add(clsName);
			//printErrorMessage("Added " + clsName + " to kb");
		}
		
	  	for(child = node.getFirstChild(); child != null; child = child.getNextSibling()) {
	  		
	  		String toplevelname = child.getNodeName();
	  	
		  	if (toplevelname.equals("own-slots")) {	

				//the class has to be created at this stage!!!
				
			  	for(child2 = child2.getNextSibling(); child2 != null; child2 = child2.getNextSibling()) {
			  		
			  		if (child2.getNodeType() != Node.ELEMENT_NODE) {
			  			continue;
			  		}
			  		
			  		String slotName = getSlotName(child2);
			  		Slot slot = getSlot(slotName);
			  		
  			  		//printErrorMessage("slot name: " + slotName);
			  		
			  		Collection slotColl = parseOwnSlot(child2);
			  		
			  		
			  		if (slotName.equals(Model.Slot.DIRECT_SUPERCLASSES)) {
			  			//iterate through list of superclasses
			  			Iterator i = slotColl.iterator();
			  			while (i.hasNext()) {
			  				cls.addDirectSuperclass((Cls) i.next());
			  			}
			  		}
			  		else if (slotName.equals(Model.Slot.DIRECT_TYPES)) {
			  			//iterate through list of superclasses
			  			Iterator i = slotColl.iterator();
			  			while (i.hasNext()) {
			  				cls.setDirectType((Cls) i.next());
			  			}
			  		}
			  		else {
					  	try {												  		
							cls.setOwnSlotValues(slot, slotColl);
						}
						catch(IllegalArgumentException iae) { 
							printErrorMessage(clsName + ": not correct type"); 
					    } 
					}										
			  	}  //end own-slot			  	
			}

		  	//overrides
		  	else if (toplevelname.equals("slot-overrides")) {

		  		for(child2 = child.getFirstChild(); child2 != null; child2 = child2.getNextSibling()) {

		  			//get the name of the slot being overriden
		  			String overriddenSlotName = getSlotName(child2);
		  			if (overriddenSlotName.equals("")) {
		  				continue;
		  			}
		  			Slot overriddenSlot = getSlot(overriddenSlotName);
		  			
		  			//go through each overridden facet
		  			for (child3 = child2.getFirstChild(); child3 != null; child3 = child3.getNextSibling()) {
						if (child3.getNodeName().equals("override")) {

					  		if (child3.getNodeType() != Node.ELEMENT_NODE) {
					  			continue;
					  		}
					  		
					  		String facetName = getAttribute(child3, "facet-name");
							Facet facet = getFacet(facetName);
							
							//printErrorMessage("facet: " + facetName);
				
					  		if (facetName.equals(Model.Facet.VALUE_TYPE)) {
								ArrayList vtArray = parseValueType(child3);
				
								Object obj = vtArray.get(0);
								ValueType type = ValueType.valueOf((String) obj);
								cls.setTemplateSlotValueType(overriddenSlot,type);
								vtArray.remove(0);
								
								if(type.equals(ValueType.CLS)) {
									cls.setTemplateSlotAllowedParents(overriddenSlot,vtArray);
								}
				
								else if(type.equals(ValueType.INSTANCE)) {
									cls.setTemplateSlotAllowedClses(overriddenSlot,vtArray);
								}
						
								else if(type.equals(ValueType.SYMBOL))
									cls.setTemplateSlotAllowedValues(overriddenSlot,vtArray);
							}													  	
							else {
								
					  			Collection facetColl = parseOwnSlot(child3);

								if (facetName.equals(Model.Facet.MAXIMUM_CARDINALITY)) {
						  			
						  			Integer infiniteMaxCard = new Integer(KnowledgeBase.MAXIMUM_CARDINALITY_UNBOUNDED);
						  			
						  			Iterator i = facetColl.iterator();
						  			Integer maxCard = (Integer) i.next();
						  			
						  			if (maxCard.equals(infiniteMaxCard)) {
						  				cls.setTemplateSlotAllowsMultipleValues(overriddenSlot, true);
						  			}
						  		}
					  			else {
								  	try {
										cls.setTemplateFacetValues(overriddenSlot, facet, facetColl);
									}
									catch(IllegalArgumentException iae) { 
										printErrorMessage(facetName + ": not correct type"); 
							   		}
							   	}		  							
						  	}						
						}
		  			}
		  		}	
		  	}
		}
	}	
	
	
	protected void parseSlot(Node node) {
		//the new class
		Slot slot;
		String slotName;

		//make the slot first
		//if the first child is not the "name" tag then something is wrong with the DOM tree
		//schema specifies sequence and order where name comes first
		
		Node child = node.getFirstChild();
		
		if (!getSlotName(child).equals(Model.Slot.NAME)) {
			//oops
			return;
		}
		else {
			Collection slotColl = parseOwnSlot(child);
			Iterator i = slotColl.iterator();
			slotName = (String) i.next();
			slot = getSlot(slotName);
		}

		for(child = child.getNextSibling(); child != null; child = child.getNextSibling()) {
	  			  		
	  		if (child.getNodeType() != Node.ELEMENT_NODE) {
	  			continue;
	  		}
	  		
	  		String ownSlotName = getSlotName(child);
	  		
	  		Slot ownslot = getSlot(ownSlotName);

	  		if (ownSlotName.equals(Model.Slot.VALUE_TYPE)) {
	  			
				ArrayList vtArray = parseValueType(child);

				Object obj = vtArray.get(0);
				ValueType type = ValueType.valueOf((String) obj);
				slot.setValueType(type);
				vtArray.remove(0);
				
				if(type.equals(ValueType.CLS)) {				
					slot.setAllowedParents(vtArray);
				}

				else if(type.equals(ValueType.INSTANCE)) {
					slot.setAllowedClses(vtArray);
				}
		
				else if(type.equals(ValueType.SYMBOL))
					slot.setAllowedValues(vtArray);
			  	}
			else {

	  			Collection slotColl = parseOwnSlot(child);
	  		
		  		if (ownSlotName.equals(Model.Slot.DIRECT_TYPES)) {
					//iterate through list of superclasses
					Iterator i = slotColl.iterator();
					while (i.hasNext()) {
						slot.setDirectType((Cls) i.next());
					}
		  		}
		  		else if (ownSlotName.equals(Model.Slot.MAXIMUM_CARDINALITY)) {
		  			
		  			Integer infiniteMaxCard = new Integer(KnowledgeBase.MAXIMUM_CARDINALITY_UNBOUNDED);
		  			
		  			Iterator i = slotColl.iterator();
		  			Integer maxCard = (Integer) i.next();
		  			
		  			if (maxCard.equals(infiniteMaxCard)) {
		  				slot.setAllowsMultipleValues(true);
		  			}
		  			else {
					  	try {
							slot.setOwnSlotValues(ownslot, slotColl);
						}
						catch(IllegalArgumentException iae) { 
							printErrorMessage(ownSlotName + ": not correct type"); 
				   		}
				   	}		  			
		  		}
		  		else {
				  	try {
						slot.setOwnSlotValues(ownslot, slotColl);
					}
					catch(IllegalArgumentException iae) { 
						printErrorMessage(ownSlotName + ": not correct type"); 
			   		}
			   	} 
		  	}

	  	}  //end own-slot
	}
	

	protected void parseInstance(Node node) {
		Instance instance;
		String instanceName = null;
		String instanceClassname = null;
		Cls instanceClass = null;
		Slot ownSlot;
		Node child, child2;
		

		
		//id must be first
	  	child = node.getFirstChild();

		for (child = node.getFirstChild(); child != null; child = child.getNextSibling()) {
			String nodeName = child.getNodeName();

			//defaults
			if (nodeName.equals("id")) {
				instanceName = getTextnodevalue(child);
			}
			else if (nodeName.equals("type"))
				instanceClassname = getTextnodevalue(child);			
			else if (nodeName.equals("own-slots")) {	
				if (!doesClassExist(instanceClassname)) {
					printErrorMessage("Can not load instance "+instanceName+" of type "+instanceClassname+":\nclass does not exist\n");
					return;
				}
				else {
					instanceClass = getClass(instanceClassname);
				}
		
				if (instanceName != null && instanceClass != null)
					instance = getInstance(instanceName, instanceClass);
				else {
					printErrorMessage("Couldn't make instance:" + instanceName);
					return;
				}
				
				for(child2 = child.getFirstChild(); child2 != null; child2 = child2.getNextSibling()) {
			  		
			  		if (child2.getNodeType() != Node.ELEMENT_NODE) {
			  			continue;
			  		}
			  		
			  		// get the slot for this own-slot
			  		String slotName = getSlotName(child2);
					if (doesSlotExist(slotName)) {
						ownSlot = getSlot(slotName);
					}
					else {
						printErrorMessage(instanceName + ": " + slotName + ": slot does not exist");
						continue;
					}
					
			  		Collection slotColl = parseOwnSlot(child2);
			  		
				  	try {
						instance.setOwnSlotValues(ownSlot, slotColl);
					}
					catch(IllegalArgumentException iae) { 
						printErrorMessage(slotName + ": not correct type"); 
				    } 
				}
			}
		}
	}
	
	private Object convertValue(ValueType type, String value) {

		Object obj = null;
				    
		if(type.equals(ValueType.STRING)) 
				obj = value;
			
		else if(type.equals(ValueType.FLOAT))
				obj = Float.valueOf(value);

		else if(type.equals(ValueType.INTEGER)) 
				obj = Integer.valueOf(value);

		else if(type.equals(ValueType.ANY))
		    	obj = value;

		else if(type.equals(ValueType.BOOLEAN))
				obj = Boolean.valueOf(value); 		

		else if(type.equals(ValueType.CLS)) {				
				obj = getClass(value);
		}

		else if(type.equals(ValueType.INSTANCE)) {
//		    	String parseInstname = parseInstanceName(value);
//				String parseInstcls = parseInstanceCls(value);
//				obj = getInstance(parseInstname, getClass(parseInstcls));
				obj = getInstance(value, null);
		}

		else if(type.equals(ValueType.SYMBOL))
			obj = value;
			
		return obj;
	}
	

	private boolean isInstance(String input) {
		if (input.startsWith("Instance"))
			return true;
		else
			return false;				
	}
	
	private boolean isClass(String input) {
		if (input.startsWith("Cls"))
			return true;
		else
			return false;				
	}
	
	private String parseInstanceName(String input) {
		String instanceName;
		int posA = input.indexOf('(');
		instanceName = input.substring(0,posA);

		return instanceName;
	}
	
	
	private String parseInstanceCls(String input) {
		String instanceCls;
		int posA = input.indexOf('(');
		int posB = input.indexOf(')', posA+1);
		instanceCls = input.substring(posA+1, posB);		
		return instanceCls;
	}
	
	//gets class name from Cls(CLASSNAME) format
	private String parseClsString(String input) {
		String cls;
		
		if (input.startsWith("Cls")) {
			int posA = input.indexOf('(');
			int posB = input.indexOf(')', posA+1);
			cls = input.substring(posA+1, posB);
		}
		else {
			cls = "";
		}
		
		return cls;
	}
	
	//gets class name from Cls(CLASSNAME) format
	private String parseSlotString(String input) {
		String slot;
		
		if (input.startsWith("Slot")) {
			int posA = input.indexOf('(');
			int posB = input.indexOf(')', posA+1);
			slot = input.substring(posA+1, posB);
		}
		else {
			slot = "";
		}
		
		return slot;
	}
	
	private Cls getClass(String clsName) {
		
		Cls cls = mainkb.getCls(clsName);
		if(cls == null){
			cls = mainkb.createCls(clsName, new Vector());
			cls.setIncluded(localIsIncluded);
		}
		
		return cls;
	}
	
//	private Slot getSlot(String slotName) {
//		Slot slot = mainkb.getSlot(slotName);
//		if (slot == null) {
//			slot = mainkb.createSlot(slotName);
//		}
//		
//		return slot;
//	}
	
	
	private boolean doesSlotExist(String name) {
		if (mainkb.getSlot(name) != null)
			return true;
		else
			return false;	
	}
	
	private boolean doesClassExist(String name) {
		if (_existingClasses.contains(name)) {
			return true;
		}
		else if (mainkb.getCls(name) != null)
			return true;
		else
			return false;
	}
	
	private Slot getSlot(String name) {
        Slot slot;
        
        Frame frame = mainkb.getFrame(name);
        if (frame == null) {
            slot = mainkb.createSlot(name);
            slot.setIncluded(localIsIncluded);
        } else if (frame instanceof Slot) {
            slot = (Slot) frame;
        } else {
 //           nameConflictError(name, "slot", frame);
            slot = getSlot(name + "_");
        }
        return slot;
    }
    
	private Facet getFacet(String facetName) {
		Facet facet = mainkb.getFacet(facetName);
		if (facet == null) {
			facet = mainkb.createFacet(facetName);
			facet.setIncluded(localIsIncluded);
		}
		
		return facet;
	}
	
	private void getOntology() {
	}
	
	
	
	/*  ERROR HANDLING STUFF FOR THIS CLASS  */
	
	private void printExtraClasses(Hashtable ht) {
		String errormsg;
		
		errormsg = "The following classes could not be added because they were subclasses of non-existant superclasses:\n";
		
		for ( Enumeration e = ht.keys() ; e.hasMoreElements();) {
				// retrieve the class name
				String thisCls = (String) e.nextElement();
				errormsg += thisCls + "\n";
		}
			
		
		//popup
		
		popupErrorMessage(errormsg);

	}
	
	
	
	/* Stuff completely stolen from Ray's Parser.java
	 */
		
	private int nInstances;
	//pass null if have null
    private Instance getInstance(String instanceName, Cls cls) {
        Instance instance = mainkb.getInstance(instanceName);
        if (instance == null) {
            instance = mainkb.createInstance(instanceName, cls, false);
            if ((++nInstances % INSTANCE_TRACE_COUNT_DELTA) == 0 && nInstances >= MIN_INSTANCE_TRACE_COUNT) {
  //              Log.trace("n=" + nInstances, this, "getInstance", instanceName, cls);
                Thread.yield();
            }
            if (cls == null) {
                itsUndefinedInstances.add(instance);
            }
            if (instance instanceof Cls) {
                itsUndefinedClses.add(instance);
            }
            
            instance.setIncluded(localIsIncluded);
	        
        } else if (instance.getDirectType() != cls && cls != null) {
            if (isTypeMismatch(instance, cls)) {
                printErrorMessage("type mismatch: " + instance + " cannot become an instance of " + cls);
            }
            Instance oldInstance = instance;
            instance = instance.setDirectType(cls);
            if ((++nInstances % INSTANCE_TRACE_COUNT_DELTA) == 0 && nInstances >= MIN_INSTANCE_TRACE_COUNT) {
  //              Log.trace("set type n=" + nInstances, this, "", instanceName, cls.getName());
                Thread.yield();
            }
            itsUndefinedInstances.remove(oldInstance);
        }	
        return instance;
    }

    private static boolean isTypeMismatch(Instance instance, Cls cls) {
        boolean isTypeMismatch;
        if (instance instanceof Cls) {
            isTypeMismatch = !cls.isClsMetaCls();
        } else if (instance instanceof Slot) {
            isTypeMismatch = !cls.isSlotMetaCls();
        } else {
            isTypeMismatch = cls.isMetaCls();
        }
        return isTypeMismatch;
    }
    
    private void checkUndefinedClasses() {
        if (!itsUndefinedClses.isEmpty()) {
            Cls standardCls = mainkb.getRootClsMetaCls();
            Cls undefinedSuperclass = getUndefinedSuperclass();
            Iterator i = itsUndefinedClses.iterator();
            while (i.hasNext()) {
                Cls cls = (Cls) i.next();
                printErrorMessage("Undefined class: " + cls.getName() + " found while " + state);
                cls.addDirectSuperclass(undefinedSuperclass);
                cls.setDirectType(standardCls);
            }
            itsUndefinedClses.clear();
        }
    }

    private void checkUndefinedInstances() {
        if (!itsUndefinedInstances.isEmpty()) {
            Cls cls = getUndefinedSuperclass();
            Iterator i = itsUndefinedInstances.iterator();
            while (i.hasNext()) {
                Instance instance = (Instance) i.next();
                printErrorMessage("Undefined instance: " + instance.getName());
                instance.setDirectType(cls);
            }
        }
    }
    
    private static final String UNDEFINED_CLASS = ":UNDEFINED";
    private Cls getUndefinedSuperclass() {
        Cls cls = mainkb.getCls(UNDEFINED_CLASS);
        if (cls == null) {
        	//sorry, fix later
//			Collection parents = CollectionUtilities.createCollection(mainkb.getRootCls());
	        cls = mainkb.createCls(UNDEFINED_CLASS, new Vector());
        }
        return cls;
    }    
}

