/*
 * The contents of this file are subject to the Mozilla Public License
 * Version 1.1 (the "License");  you may not use this file except in compliance
 * with the License.  You may obtain a copy of the License at
 * http://www.mozilla.org/MPL/
 *
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License for
 * the specific language governing rights and limitations under the License.
 *
 * The Original Code is Protege-2000.
 *
 * The Initial Developer of the Original Code is Stanford University. Portions
 * created by Stanford University are Copyright (C) 2001.  All Rights Reserved.
 *
 * Protege-2000 was developed by Stanford Medical Informatics
 * (http://www.smi.stanford.edu) at the Stanford University School of Medicine
 * with support from the National Library of Medicine, the National Science
 * Foundation, and the Defense Advanced Research Projects Agency.  Current
 * information about Protege can be obtained at http://protege.stanford.edu
 *
 * This code developed by The Department of Biomedical and Health Informatics at 
 * at the University of Washington (http://www.bhi.washington.edu)
 *
 */
 
package edu.washington.bhi.protegex.storage.xml;

import org.w3c.dom.*;

import javax.xml.parsers.DocumentBuilderFactory;
import javax.xml.parsers.DocumentBuilder;

import edu.stanford.smi.protege.model.*;

/* Writes out XML from Protege
 */
public class XMLCreator extends XMLBase{


	//variables	
	protected Document doc;
	
	//root node for the DOM tree
	protected Element main;

	//initialize function
	public XMLCreator()
	{
		createDocument();
	}
	
	protected void createDocument() {
		try
        {
            //Create Document
            DocumentBuilderFactory dbf = DocumentBuilderFactory.newInstance();
            DocumentBuilder db = dbf.newDocumentBuilder();
            doc = db.newDocument();
        }
       	
       	catch(Exception e)
        {
         
         popupErrorMessage("Error " + e);
            
        }
	}

    protected Element addChildElement(Element parent, String name, String value){
		Element element = doc.createElement(normalize(name));
		element.appendChild(doc.createTextNode(normalize(value)));
		parent.appendChild(element);
		parent.appendChild(doc.createTextNode("\n"));
		return element;
    }
    
    protected Element addChildtoFront(Element parent, String name, String value) {
    	Element element = doc.createElement(normalize(name));
		element.appendChild(doc.createTextNode(normalize(value)));
				
  		//get the first element
  		Node firstnode = parent.getFirstChild();
  		
  		//stick new node at the beginning
  		parent.insertBefore(element, firstnode);
  		parent.insertBefore(doc.createTextNode("\n"), firstnode);
  		
  		return element;	
    }
    
    protected Element addRootElement(Element parent, String name){
		Element element = doc.createElement(normalize(name));
		parent.appendChild(element);
		parent.appendChild(doc.createTextNode("\n"));
		return element;
    }	
    
    
    protected void addElementtoMain(Element element) {
    	main.appendChild(element);
	    main.appendChild(doc.createTextNode("\n"));
    }

    protected void deleteNode(Element parent, Element child) {
		parent.removeChild(child);
    }

    protected void addAttribute(Element parent, String name, String value){
		Attr attr = doc.createAttribute(normalize(name));
		attr.setValue(normalize(value));
		parent.setAttributeNode(attr);
    }
    
    protected void initializeMain(String value) {
    	main = doc.createElement(value);
        main.appendChild(doc.createTextNode("\n"));
    }
    
    protected void addMaintoDoc() {
    	doc.appendChild(main);
    }
    
    protected ValueType getObjectValueType(Object obj) {
		if (obj instanceof Boolean)
        	return ValueType.BOOLEAN;
        else if (obj instanceof Cls)
        	return ValueType.CLS;
        else if (obj instanceof Float)
        	return ValueType.FLOAT;
        else if (obj instanceof Instance)
        	return ValueType.INSTANCE;
        else if (obj instanceof Integer)
        	return ValueType.INTEGER;
        else if (obj instanceof String)
        	return ValueType.STRING;
        else
        	return ValueType.ANY;        
    }
    
    protected String convertObjecttoString(Object obj, ValueType type) {
    	if(type.equals(ValueType.STRING))
		    return obj.toString();
		else if(type.equals(ValueType.FLOAT))
		    return obj.toString();
		else if(type.equals(ValueType.INTEGER))
		    return obj.toString();
		else if(type.equals(ValueType.ANY))
		    return obj.toString();
		else if(type.equals(ValueType.BOOLEAN))
		    return obj.toString();
		else if(type.equals(ValueType.CLS)) {
		    Cls cls = (Cls) obj;
		    return cls.getName();
		}
		else if(type.equals(ValueType.INSTANCE)) {
			Instance inst = (Instance) obj;
			return inst.getName();
//			return inst.getName()+"("+inst.getDirectType().getName()+")";
		}
		else if(type.equals(ValueType.SYMBOL))
		    return obj.toString();
		else return "Unknown";
    }
}